from logging import getLogger
from typing import Callable, Iterable, TYPE_CHECKING

from ..config import CONFIG as _CONFIG
from ..entity import Entity
from .result import SearchResultList
from .annotation import Annotations, Annotation, Correctness, Completeness

if TYPE_CHECKING:  # avoid circular dependency
    from ..index import Result


LOGGER = getLogger(__name__)


RelevanceCriterion = Callable[[Annotation], bool]


def default_relevance_criterion(annotation: Annotation) -> bool:
    if annotation.correctness == Correctness.FULL and annotation.completeness == Completeness.EXACT_OVERLAP:
        return True  # perfect match
    elif annotation.correctness == Correctness.PARTIAL and annotation.completeness == Completeness.EXACT_OVERLAP:
        return True  # different entity
    elif annotation.note == 'navíc pouze interpunknce':
        return True  # punctuation around entity
    return False


class Evaluation:
    CONFIG = _CONFIG['search.Evaluation']
    BETA = CONFIG.getfloat('f_score_beta')
    TOPN = CONFIG.getint('number_of_results')

    def __init__(self, result_list: SearchResultList, annotations: Annotations,
                 entities: Iterable[Entity],
                 relevance_criterion: RelevanceCriterion = default_relevance_criterion):
        self.result_list = result_list
        self.annotations = annotations.annotations
        self.entities = list(entities)
        self.relevance_criterion = relevance_criterion

    @property
    def number_of_total_relevant_results(self) -> int:
        number_of_total_relevant_results = 0
        for entity in self.entities:
            for annotation in self.annotations[entity].values():
                if self.relevance_criterion(annotation):
                    number_of_total_relevant_results += 1
        return number_of_total_relevant_results

    @property
    def number_of_found_relevant_results(self) -> int:
        number_of_found_relevant_results = 0
        for entity in self.entities:
            results = self._get_topn_results(entity)
            for (result_head, _) in results:
                if result_head not in self.annotations[entity]:
                    continue
                annotation = self.annotations[entity][result_head]
                if self.relevance_criterion(annotation):
                    number_of_found_relevant_results += 1
        return number_of_found_relevant_results

    @property
    def number_of_found_unannotated_results(self) -> int:
        number_of_found_unannotated_results = 0
        for entity in self.entities:
            results = self._get_topn_results(entity)
            for (result_head, _) in results:
                if result_head not in self.annotations[entity]:
                    number_of_found_unannotated_results += 1
        return number_of_found_unannotated_results

    @property
    def number_of_found_results(self) -> int:
        number_of_found_results = 0
        for entity in self.entities:
            results = self._get_topn_results(entity)
            for (result_head, _) in results:
                number_of_found_results += 1
        return number_of_found_results

    def _get_topn_results(self, entity: Entity) -> Iterable['Result']:
        for result in self.result_list[entity][:self.TOPN]:
            yield result

    @property
    def precision(self) -> float:
        if self.number_of_found_unannotated_results > 0:
            message = '{} results have not been annotated; treating them as non-relevant'.format(
                self.number_of_found_unannotated_results)
            LOGGER.warning(message)
        if self.number_of_found_results == 0:
            if self.number_of_total_relevant_results > 0:
                return 0.0
            else:
                raise ValueError('There are no results in the result list and no known relevant results')
        precision = 1.0 * self.number_of_found_relevant_results / self.number_of_found_results
        return precision

    @property
    def recall(self) -> float:
        if self.number_of_found_unannotated_results > 0:
            message = '{} results have not been annotated; treating them as non-relevant'.format(
                self.number_of_found_unannotated_results)
            LOGGER.warning(message)
        if self.number_of_total_relevant_results == 0:
            return 1.0
        recall = 1.0 * self.number_of_found_relevant_results / self.number_of_total_relevant_results
        return recall

    @property
    def f_score(self) -> float:
        beta_squared = self.BETA ** 2
        numerator = (1 + beta_squared) * self.precision * self.recall
        denominator = (beta_squared * self.precision) + self.recall
        f_score = numerator / denominator if denominator > 0.0 else 0.0
        return f_score

    def __str__(self) -> str:
        summary = 'The results achieved Precision@{topn} of {:.2f}%, Recall@{topn} of {:.2f}%, and F@{topn}-score of {:.2f}%.'
        summary = summary.format(100.0 * self.precision, 100.0 * self.recall, 100.0 * self.f_score, topn=self.TOPN)
        return summary

    def __repr__(self) -> str:
        return str(self)
