from collections.abc import Sequence, Callable
import gc
import multiprocessing.pool
from typing import Set, List, Tuple
from functools import lru_cache
import sys
import os

import regex
from regex import sub
from edit_distance import SequenceMatcher
from gensim.utils import deaccent
import torch

from .config import CONFIG as _CONFIG


CONFIG = _CONFIG['util']

TOKEN = regex.compile(r'\w+')


def cache(user_function: Callable) -> Callable:
    maxsize = CONFIG.getint('cache_maxsize')
    return lru_cache(maxsize=maxsize)(user_function)


def collect_garbage_from_torch() -> None:
    gc.collect()
    torch.cuda.empty_cache()


def suppress_stderr(func):
    def wrapper(*args, **kwargs):
        stderr_tmp = sys.stderr
        null = open(os.devnull, 'w')
        sys.stderr = null
        try:
            result = func(*args, **kwargs)
            sys.stderr = stderr_tmp
            return result
        except:  # noqa:E722
            sys.stderr = stderr_tmp
            raise
    return wrapper


def lossless_strip(text: str) -> Tuple[str, str, str]:
    strip_regex = r'^(?P<left_stripped_bits>\P{L}*)(?P<body>.*?)(?P<right_stripped_bits>\P{L}*)$'
    match = regex.match(strip_regex, text)
    assert match
    left_stripped_bits = match.group('left_stripped_bits')
    body = match.group('body')
    right_stripped_bits = match.group('right_stripped_bits')
    return (left_stripped_bits, body, right_stripped_bits)


def strip(text: str) -> str:
    left_stripped_bits, body, right_stripped_bits = lossless_strip(text)
    return body


def normalize_ocr(text: str) -> str:
    text = sub(r'-\s*\n\s*', '', text)
    return ' '.join(text.split())


def tokenize(text: str) -> List[str]:
    tokens = [match.group() for match in TOKEN.finditer(text)]
    return tokens


def preprocess(text: str) -> List[str]:
    text = normalize_ocr(text)
    text = deaccent(text)
    tokens = tokenize(text)
    return tokens


def normalize(text: str) -> str:
    return ' '.join(preprocess(text))


def error_rate(first: Sequence, second: Sequence) -> float:
    max_distance = max(len(first), len(second))
    if max_distance == 0.0:
        return 1.0
    distance = SequenceMatcher(a=first, b=second).distance()
    error_rate = 1.0 * distance / max_distance
    return error_rate


def extract_ngrams(text: str, minn: int, maxn: int) -> Set[str]:
    ngrams = set()
    maxn = min(maxn, len(text))
    for i in range(minn, maxn+1):
        for j in range(len(text)-i+1):
            ngram = text[j:j+i]
            ngrams.add(ngram)
    return ngrams


def jaccard_index(first: Set, second: Set) -> float:
    union = len(first | second)
    if union == 0:
        return 1.0
    intersection = len(first & second)
    similarity = 1.0 * intersection / (union if union > 0.0 else 1.0)
    return similarity


class NoDaemonProcess(multiprocessing.Process):
    @property
    def daemon(self):
        return False

    @daemon.setter
    def daemon(self, value):
        pass


class NoDaemonForkingContext(type(multiprocessing.get_context('fork'))):
    Process = NoDaemonProcess


class NestableForkingPool(multiprocessing.pool.Pool):
    def __init__(self, *args, **kwargs):
        kwargs['context'] = NoDaemonForkingContext()
        super().__init__(*args, **kwargs)
